from __future__ import annotations

from collections.abc import Callable, Coroutine
from inspect import iscoroutinefunction
from typing import Any, Protocol, TypeAlias, overload

from .typing import AsyncComponent, Component, Context, SyncComponent
from .typing import T as TProps
from .typing import U as TSelf

# -- Typing for "full" function components and context only method components.

_SyncFunctionComponent: TypeAlias = Callable[[TProps, Context], Component]
"""
Protocol definition for sync function components that have both a properties and a context argument.
"""

_AsyncFunctionComponent: TypeAlias = Callable[[TProps, Context], Coroutine[Any, Any, Component]]
"""
Protocol definition for async function components that have both a properties and a context argument.
"""

# -- Typing for context-only function components.

_ContextOnlySyncFunctionComponent: TypeAlias = Callable[[Context], Component]
"""
Protocol definition for sync function components that only have a context argument.
"""


class _DecoratedContextOnlySyncFunctionComponent(SyncComponent, Protocol):
    """
    Protocol definition for sync components that are also callable, and return a sync
    component when called.
    """

    def __call__(self) -> SyncComponent: ...


_ContextOnlyAsyncFunctionComponent: TypeAlias = Callable[[Context], Coroutine[Any, Any, Component]]
"""
Protocol definition for async function components that only have a context argument.
"""


class _DecoratedContextOnlyAsyncFunctionComponent(SyncComponent, Protocol):
    """
    Protocol definition for async components that are also callable, and return an async
    component when called.
    """

    def __call__(self) -> SyncComponent: ...


# -- Typing for "full" method components.

_SyncMethodComponent: TypeAlias = Callable[[TSelf, TProps, Context], Component]
"""
Protocol definition for sync method components that have both a properties and a context argument.
"""

_AsyncMethodComponent: TypeAlias = Callable[[TSelf, TProps, Context], Coroutine[Any, Any, Component]]
"""
Protocol definition for async method components that have both a properties and a context argument.
"""


# -- Component decorators.


class ComponentDecorators:
    """
    Function component decorators.
    """

    __slots__ = ()

    # -- Function component decorator.

    @overload
    def __call__(self, func: _SyncFunctionComponent[TProps]) -> Callable[[TProps], SyncComponent]: ...

    @overload
    def __call__(self, func: _AsyncFunctionComponent[TProps]) -> Callable[[TProps], AsyncComponent]: ...

    def __call__(
        self,
        func: _SyncFunctionComponent[TProps] | _AsyncFunctionComponent[TProps],
    ) -> Callable[[TProps], SyncComponent] | Callable[[TProps], AsyncComponent]:
        """
        Decorator that converts the decorated function into one that must be called with
        the function component's properties and returns a component instance.

        If used on an async function, the resulting component will also be async;
        otherwise it will be sync.

        Example:

        ```python
        @component
        def my_component(props: int, context: Context) -> Component:
            return html.p(f"Value: {props}")

        async def render() -> str:
           return await Renderer().render(
               my_component(42)
           )
        ```

        Arguments:
            func: The decorated function.

        Returns:
            A function that must be called with the function component's properties and
            returns a component instance. (Or loosly speaking, an `HTMYComponentType` which
            can be "instantiated" with the function component's properties.)
        """

        if iscoroutinefunction(func):

            def async_wrapper(props: TProps) -> AsyncComponent:
                # This function must be async, in case the renderer inspects it to decide how to handle it.
                async def component(context: Context) -> Component:
                    return await func(props, context)  # type: ignore[no-any-return]

                component.htmy = component  # type: ignore[attr-defined]
                return component  # type: ignore[return-value]

            return async_wrapper
        else:

            def sync_wrapper(props: TProps) -> SyncComponent:
                def component(context: Context) -> Component:
                    return func(props, context)  # type: ignore[return-value]

                component.htmy = component  # type: ignore[attr-defined]
                return component  # type: ignore[return-value]

            return sync_wrapper

    @overload
    def function(self, func: _SyncFunctionComponent[TProps]) -> Callable[[TProps], SyncComponent]: ...

    @overload
    def function(self, func: _AsyncFunctionComponent[TProps]) -> Callable[[TProps], AsyncComponent]: ...

    def function(
        self,
        func: _SyncFunctionComponent[TProps] | _AsyncFunctionComponent[TProps],
    ) -> Callable[[TProps], SyncComponent] | Callable[[TProps], AsyncComponent]:
        """
        Decorator that converts the decorated function into one that must be called with
        the function component's properties and returns a component instance.

        If used on an async function, the resulting component will also be async;
        otherwise it will be sync.

        This function is just an alias for `__call__()`.

        Example:

        ```python
        @component.function
        def my_component(props: int, context: Context) -> Component:
            return html.p(f"Value: {props}")

        async def render() -> str:
           return await Renderer().render(
               my_component(42)
           )

        Arguments:
            func: The decorated function.

        Returns:
            A function that must be called with the function component's properties and
            returns a component instance. (Or loosly speaking, an `HTMYComponentType` which
            can be "instantiated" with the function component's properties.)
        """
        return self(func)

    # -- Context-only function component decorator.

    @overload
    def context_only(
        self, func: _ContextOnlySyncFunctionComponent
    ) -> _DecoratedContextOnlySyncFunctionComponent: ...

    @overload
    def context_only(
        self, func: _ContextOnlyAsyncFunctionComponent
    ) -> _DecoratedContextOnlyAsyncFunctionComponent: ...

    def context_only(
        self,
        func: _ContextOnlySyncFunctionComponent | _ContextOnlyAsyncFunctionComponent,
    ) -> _DecoratedContextOnlySyncFunctionComponent | _DecoratedContextOnlyAsyncFunctionComponent:
        """
        Decorator that converts the decorated function into a component.

        If used on an async function, the resulting component will also be async;
        otherwise it will be sync.

        Example:

        ```python
        @component.context_only
        def my_component(ctx):
            return "Context only function component."

        async def render() -> str:
           return await Renderer().render(
               my_component()
           )
        ```

        Arguments:
            func: The decorated function.

        Returns:
            The created component.
        """

        def wrapper() -> SyncComponent | AsyncComponent:
            func.htmy = func  # type: ignore[union-attr]
            return func  # type: ignore[return-value]

        # This assignment adds support for context-only function components without call signature.
        wrapper.htmy = func  # type: ignore[attr-defined]
        return wrapper  # type: ignore[return-value]

    # -- Method component decorator.

    @overload
    def method(
        self, func: _SyncMethodComponent[TSelf, TProps]
    ) -> Callable[[TSelf, TProps], SyncComponent]: ...

    @overload
    def method(
        self, func: _AsyncMethodComponent[TSelf, TProps]
    ) -> Callable[[TSelf, TProps], AsyncComponent]: ...

    def method(
        self,
        func: _SyncMethodComponent[TSelf, TProps] | _AsyncMethodComponent[TSelf, TProps],
    ) -> Callable[[TSelf, TProps], SyncComponent] | Callable[[TSelf, TProps], AsyncComponent]:
        """
        Decorator that converts the decorated method into one that must be called with
        the method component's properties and returns a component instance.

        If used on an async method, the resulting component will also be async;
        otherwise it will be sync.

        Example:

        ```python
        @dataclass
        class MyBusinessObject:
            message: str

            @component.method
            def paragraph(self, props: int, context: Context) -> Component:
                return html.p(f"{self.message} {props}")


        async def render() -> str:
            return await Renderer().render(
                MyBusinessObject("Hi!").paragraph(42)
            )
        ```

        Arguments:
            func: The decorated method.

        Returns:
            A method that must be called with the method component's properties and
            returns a component instance. (Or loosly speaking, an `HTMYComponentType` which
            can be "instantiated" with the method component's properties.)
        """
        if iscoroutinefunction(func):

            def async_wrapper(self: TSelf, props: TProps) -> AsyncComponent:
                # This function must be async, in case the renderer inspects it to decide how to handle it.
                async def component(context: Context) -> Component:
                    return await func(self, props, context)  # type: ignore[no-any-return]

                component.htmy = component  # type: ignore[attr-defined]
                return component  # type: ignore[return-value]

            return async_wrapper
        else:

            def sync_wrapper(self: TSelf, props: TProps) -> SyncComponent:
                def component(context: Context) -> Component:
                    return func(self, props, context)  # type: ignore[return-value]

                component.htmy = component  # type: ignore[attr-defined]
                return component  # type: ignore[return-value]

            return sync_wrapper

    # -- Context-only function component decorator.

    @overload
    def context_only_method(
        self, func: _SyncFunctionComponent[TSelf]
    ) -> Callable[[TSelf], SyncComponent]: ...

    @overload
    def context_only_method(
        self, func: _AsyncFunctionComponent[TSelf]
    ) -> Callable[[TSelf], AsyncComponent]: ...

    def context_only_method(
        self,
        func: _SyncFunctionComponent[TSelf] | _AsyncFunctionComponent[TSelf],
    ) -> Callable[[TSelf], SyncComponent] | Callable[[TSelf], AsyncComponent]:
        """
        Decorator that converts the decorated method into one that must be called
        without any arguments and returns a component instance.

        If used on an async method, the resulting component will also be async;
        otherwise it will be sync.

        Example:

        ```python
        @dataclass
        class MyBusinessObject:
            message: str

            @component.context_only_method
            def paragraph(self, context: Context) -> Component:
                return html.p(f"{self.message} Goodbye!")


        async def render() -> str:
            return await Renderer().render(
                MyBusinessObject("Hello!").paragraph()
            )
        ```

        Arguments:
            func: The decorated method.

        Returns:
            A method that must be called without any arguments and returns a component instance.
            (Or loosly speaking, an `HTMYComponentType` which can be "instantiated" by calling
            the method.)
        """
        # A context only method component must be implemented in the same way as
        # a function component. The self argument replaces the props argument
        # and it is added automatically by Python when the method is called.
        # Even the type hint must be the same.
        # This implementation doesn't make the function itself a component though,
        # so the call signature is always necessary (unlike for context-only function
        # components).
        return self(func)


component = ComponentDecorators()
"""
Decorators for converting functions into components

This is an instance of `ComponentDecorators`.
"""



__version__ = "0.10.2"

from .core import ContextAware as ContextAware
from .core import Formatter as Formatter
from .core import Fragment as Fragment
from .core import SafeStr as SafeStr
from .core import SkipProperty as SkipProperty
from .core import Text as Text
from .core import WithContext as WithContext
from .core import XBool as XBool
from .core import xml_format_string as xml_format_string
from .error_boundary import ErrorBoundary as ErrorBoundary
from .function_component import component as component
from .renderer import Renderer as Renderer
from .renderer import StreamingRenderer as StreamingRenderer
from .renderer.typing import RendererType as RendererType
from .snippet import Slots as Slots
from .snippet import Snippet as Snippet
from .tag import Tag as Tag
from .tag import TagWithProps as TagWithProps
from .tag import wildcard_tag as wildcard_tag
from .typing import AsyncComponent as AsyncComponent
from .typing import AsyncContextProvider as AsyncContextProvider
from .typing import Component as Component
from .typing import ComponentSequence as ComponentSequence
from .typing import ComponentType as ComponentType
from .typing import Context as Context
from .typing import ContextKey as ContextKey
from .typing import ContextProvider as ContextProvider
from .typing import ContextValue as ContextValue
from .typing import HTMYComponentType as HTMYComponentType
from .typing import MutableContext as MutableContext
from .typing import Properties as Properties
from .typing import PropertyValue as PropertyValue
from .typing import StrictComponentType as StrictComponentType
from .typing import SyncComponent as SyncComponent
from .typing import SyncContextProvider as SyncContextProvider
from .utils import as_component_sequence as as_component_sequence
from .utils import as_component_type as as_component_type
from .utils import is_component_sequence as is_component_sequence
from .utils import is_component_type as is_component_type
from .utils import is_htmy_component_type as is_htmy_component_type
from .utils import join
from .utils import join_components as join_components

join_classes = join

HTMY = Renderer
"""Deprecated alias for `Renderer`."""



from .core import MD as MD
from .core import MarkdownParser as MarkdownParser
from .typing import MarkdownMetadataDict as MarkdownMetadataDict
from .typing import MarkdownParserFunction as MarkdownParserFunction
from .typing import MarkdownRenderFunction as MarkdownRenderFunction
from .typing import ParsedMarkdown as ParsedMarkdown



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar
from xml.sax.saxutils import unescape

try:
    from lxml.etree import _Element as Element
    from lxml.etree import tostring as etree_to_string
    from lxml.html import fragment_fromstring as etree_from_string
except ImportError:
    from xml.etree.ElementTree import Element  # type: ignore[assignment]
    from xml.etree.ElementTree import fromstring as etree_from_string  # type: ignore[assignment]
    from xml.etree.ElementTree import tostring as etree_to_string  # type: ignore[no-redef]

from .core import Fragment, SafeStr
from .tag import wildcard_tag
from .typing import ComponentType, Properties

if TYPE_CHECKING:
    from collections.abc import Callable, Generator, Mapping


class ETreeConverter:
    """
    Utility for converting XML strings to custom components.

    By default the converter uses the standard library's `xml.etree.ElementTree`
    module for string to element tree, and element tree to string conversion,
    but if `lxml` is installed, it will be used instead.

    Installing `lxml` is recommended for better performance and additional features,
    like performance and support for broken HTML fragments. **Important:** `lxml` is
    far more lenient and flexible than the standard library, so having it installed is
    not only a performance boost, but it may also slightly change the element conversion
    behavior in certain edge-cases!
    """

    __slots__ = ("_rules",)

    _htmy_fragment: ClassVar[str] = "htmy_fragment"
    """
    Placeholder tag name that's used to wrap possibly multi-root XML snippets into a valid
    XML document with a single root that can be processed by standard tools.
    """

    def __init__(self, rules: Mapping[str, Callable[..., ComponentType]]) -> None:
        """
        Initialization.

        Arguments:
            rules: Tag-name to component conversion rules.
        """
        self._rules = rules

    def convert(self, element: str) -> ComponentType:
        """Converts the given (possibly multi-root) XML string to a component."""
        if len(self._rules) == 0:
            return SafeStr(element)

        element = f"<{self._htmy_fragment}>{element}</{self._htmy_fragment}>"
        return self.convert_element(etree_from_string(element))  # noqa: S314 # Only use XML strings from a trusted source.

    def convert_element(self, element: Element) -> ComponentType:
        """Converts the given `Element` to a component."""
        rules = self._rules
        if len(rules) == 0:
            return SafeStr(etree_to_string(element, encoding="unicode"))

        tag: str = element.tag  # type: ignore[assignment]
        component = Fragment if tag == self._htmy_fragment else rules.get(tag)
        children = self._convert_children(element)
        properties = self._convert_properties(element)

        return (
            wildcard_tag(*children, htmy_name=tag, **properties)
            if component is None
            else component(
                *children,
                **properties,
            )
        )

    def _convert_properties(self, element: Element) -> Properties:
        """
        Converts the attributes of the given `Element` to a `Properties` mapping.

        This method should not alter property names in any way.
        """
        return {key: unescape(value) for key, value in element.items()}

    def _convert_children(self, element: Element) -> Generator[ComponentType, None, None]:
        """
        Generator that converts all (text and `Element`) children of the given `Element` to a component.
        """
        if text := self._process_text(element.text):
            yield text

        for child in element:
            yield self.convert_element(child)
            if tail := self._process_text(child.tail):
                yield tail

    def _process_text(self, escaped_text: str | None) -> str | None:
        """Processes a single XML-escaped text child."""
        return unescape(escaped_text) if escaped_text else None



from .baseline import Renderer as _BaselineRenderer
from .default import Renderer as _DefaultRenderer
from .typing import is_renderer as is_renderer
from .typing import is_streaming_renderer as is_streaming_renderer

Renderer = _DefaultRenderer
"""The default renderer."""

StreamingRenderer = _BaselineRenderer
"""The default streaming renderer."""

BaselineRenderer = _BaselineRenderer
"""The baseline renderer."""



from __future__ import annotations

from typing import TYPE_CHECKING

from .core import Formatter, SafeStr
from .utils import join_components

if TYPE_CHECKING:
    from .typing import Component, ComponentSequence, ComponentType, Context, Properties, PropertyValue


class _TagWithPropsImpl:
    __slots__ = ("name", "props")

    def __init__(self, name: str, props: Properties) -> None:
        self.name = name
        self.props = props

    def htmy(self, context: Context) -> ComponentType:
        formatter: Formatter = context.get(Formatter, _default_formatter)
        return SafeStr(f"<{self.name} {' '.join(formatter.format(n, v) for n, v in self.props.items())}/>")


class _TagImpl:
    __slots__ = ("child_separator", "children", "name", "props")

    def __init__(
        self,
        name: str,
        props: Properties,
        children: ComponentSequence,
        child_separator: ComponentType,
    ) -> None:
        self.name = name
        self.props = props
        self.children = children
        self.child_separator = child_separator

    def htmy(self, context: Context) -> Component:
        formatter: Formatter = context.get(Formatter, _default_formatter)
        name = self.name
        return (
            SafeStr(f"<{name} {' '.join(formatter.format(n, v) for n, v in self.props.items())}>"),
            *(
                self.children
                if self.child_separator is None
                else join_components(self.children, separator=self.child_separator, pad=True)
            ),
            SafeStr(f"</{name}>"),
        )


_default_formatter = Formatter()


class TagWithProps:
    """
    Creates a tag type that can have properties, but not children.

    Arguments:
        **props: Tag properties.
    """

    __slots__ = ("name",)

    def __init__(self, name: str) -> None:
        self.name = name

    def __call__(self, **props: PropertyValue) -> ComponentType:
        return _TagWithPropsImpl(self.name, props)


class Tag:
    __slots__ = ("child_separator", "name")

    def __init__(self, name: str, child_separator: ComponentType = "\n") -> None:
        self.name = name
        self.child_separator = child_separator

    def __call__(self, *children: ComponentType, **props: PropertyValue) -> ComponentType:
        return _TagImpl(self.name, props, children, self.child_separator)


def wildcard_tag(
    *children: ComponentType,
    htmy_name: str,
    htmy_child_separator: ComponentType = None,
    **props: PropertyValue,
) -> ComponentType:
    """
    Creates a tag that can have both children and properties, and whose tag name and
    child separator can be set.

    Arguments:
        *children: Children components.
        htmy_name: The tag name to use for this tag.
        htmy_child_separator: Optional separator component to add between children.
        **props: Tag properties.
    """
    return _TagImpl(htmy_name, props, children, htmy_child_separator)



import json
from collections.abc import Mapping
from pathlib import Path
from typing import Any, ClassVar, overload

from async_lru import alru_cache

from .core import ContextAware
from .io import open_file

TranslationResource = Mapping[str, Any]
"""Translation resource type."""


class I18nError(Exception): ...


class I18nKeyError(I18nError): ...


class I18nValueError(I18nError): ...


class I18n(ContextAware):
    """
    Context-aware async internationalization utility.
    """

    __slots__ = ("_fallback", "_path")

    _root_keys: ClassVar[frozenset[str]] = frozenset(("", "."))
    """Special keys that represent the "root" object, i.e. the entire translation resource file."""

    def __init__(self, path: str | Path, fallback: str | Path | None = None) -> None:
        """
        Initialization.

        Arguments:
            path: Path to the root directory that contains the translation resources.
            fallback: Optional fallback path to use if `path` doesn't contain the required resources.
        """
        self._path: Path = Path(path) if isinstance(path, str) else path
        self._fallback: Path | None = Path(fallback) if isinstance(fallback, str) else fallback

    @overload
    async def get(self, dotted_path: str, key: str) -> Any: ...

    @overload
    async def get(self, dotted_path: str, key: str, **kwargs: Any) -> str: ...

    async def get(self, dotted_path: str, key: str, **kwargs: Any) -> Any:
        """
        Returns the translation resource at the given location.

        If keyword arguments are provided, it's expected that the referenced data
        is a [format string](https://docs.python.org/3/library/string.html#formatstrings)
        which can be fully formatted using the given keyword arguments.

        Arguments:
            dotted_path: A package-like (dot separated) path to the file that contains
                the required translation resource, relative to `path`.
            key: The key in the translation resource whose value is requested. Use dots to reference
                embedded attributes.

        Returns:
            The loaded value.

        Raises:
            I18nError: If the given translation resource is not found or invalid.
        """
        try:
            return await self._resolve(self._path, dotted_path, key, **kwargs)
        except I18nError:
            if self._fallback is None:
                raise

            return await self._resolve(self._fallback, dotted_path, key, **kwargs)

    @classmethod
    async def _resolve(cls, root: Path, dotted_subpath: str, key: str, **kwargs: Any) -> Any:
        """
        Resolves the given translation resource.

        Arguments:
            root: The root path to use.
            dotted_subpath: Subpath under `root` with dots as separators.
            key: The key in the translation resource.

        Returns:
            The resolved translation resource.

        Raises:
            I18nKeyError: If the translation resource doesn't contain the requested key.
            I18nValueError: If the translation resource is not found or its content is invalid.
        """
        result = await load_translation_resource(resolve_json_path(root, dotted_subpath))
        if key in cls._root_keys:
            return result

        for k in key.split("."):
            try:
                result = result[k]
            except KeyError as e:
                raise I18nKeyError(f"Key not found: {key}") from e

        if len(kwargs) > 0:
            if not isinstance(result, str):
                raise I18nValueError("Formatting is only supported for strings.")

            return result.format(**kwargs)

        return result


@alru_cache()
async def load_translation_resource(path: Path) -> TranslationResource:
    """
    Loads the translation resource from the given path.

    Arguments:
        path: The path of the translation resource to load.

    Returns:
        The loaded translation resource.

    Raises:
        I18nValueError: If the translation resource is not a JSON dict.
    """

    try:
        async with await open_file(path, "r") as f:
            content = await f.read()
    except FileNotFoundError as e:
        raise I18nValueError(f"Translation resource not found: {str(path)}") from e

    try:
        result = json.loads(content)
    except json.JSONDecodeError as e:
        raise I18nValueError("Translation resource decoding failed.") from e

    if isinstance(result, dict):
        return result

    raise I18nValueError("Only dict translation resources are allowed.")


def resolve_json_path(root: Path, dotted_subpath: str) -> Path:
    """
    Resolves the given dotted subpath relative to root.

    Arguments:
        root: The root path.
        dotted_subpath: Subpath under `root` with dots as separators.

    Returns:
        The resolved path.

    Raises:
        I18nValueError: If the given dotted path is invalid.
    """
    *dirs, name = dotted_subpath.split(".")
    if not name:
        raise I18nValueError("Invalid path.")

    return root / Path(*dirs) / f"{name}.json"



from __future__ import annotations

from typing import TYPE_CHECKING

from anyio import open_file as open_file

if TYPE_CHECKING:
    from pathlib import Path


async def load_text_file(path: str | Path) -> str:
    """Loads the text content from the given path."""
    async with await open_file(path, "r") as f:
        return await f.read()



from __future__ import annotations

from collections.abc import Generator
from typing import TYPE_CHECKING, Any, TypeGuard

from htmy.typing import ComponentType

if TYPE_CHECKING:
    from .typing import Component, ComponentSequence, ComponentType, HTMYComponentType


def join_components(
    components: ComponentSequence,
    separator: ComponentType,
    pad: bool = False,
) -> Generator[ComponentType, None, None]:
    """
    Joins the given components using the given separator.

    Arguments:
        components: The components to join.
        separator: The separator to use.
        pad: Whether to add a separator before the first and after the last components.
    """
    if len(components) == 0:
        return

    if pad:
        yield separator

    components_iterator = iter(components)
    yield next(components_iterator)

    for component in components_iterator:
        yield separator
        yield component

    if pad:
        yield separator


def join(*items: str | None, separator: str = " ") -> str:
    """
    Joins the given strings with the given separator, skipping `None` values.
    """
    return separator.join(i for i in items if i)


def is_htmy_component_type(comp: Any) -> TypeGuard[HTMYComponentType]:
    """Returns whether the given object is a `HTMYComponentType` (has a `htmy` method)."""
    # Lazy check, should be good enough.
    return hasattr(comp, "htmy")


def is_component_type(comp: Any) -> TypeGuard[ComponentType]:
    """Returns whether the given object is a `ComponentType`."""
    # Lazy check when it comes the htmy components, should be good enough.
    # No `is_htmy_component_type()` call to avoid trivial extra function call.
    return comp is None or isinstance(comp, str) or hasattr(comp, "htmy")


def is_component_sequence(comp: Component) -> TypeGuard[ComponentSequence]:
    """Returns whether the given component is a `ComponentSequence`."""
    return isinstance(comp, (list, tuple))


def as_component_sequence(comp: Component) -> ComponentSequence:
    """Returns the given component as a `ComponentSequence`."""
    # mypy doesn't understand the `is_component_sequence` type guard.
    return comp if is_component_sequence(comp) else (comp,)  # type: ignore[return-value]


def as_component_type(comp: Component) -> ComponentType:
    """Returns the given component as a `ComponentType` (not sequence)."""
    from .core import Fragment

    # mypy doesn't understand the `is_component_sequence` type guard.
    return comp if not is_component_sequence(comp) else Fragment(*comp)  # type: ignore[return-value]



from __future__ import annotations

from asyncio import gather as asyncio_gather
from collections import ChainMap, deque
from inspect import isawaitable, iscoroutinefunction
from typing import TYPE_CHECKING, TypeAlias

from htmy.core import xml_format_string
from htmy.typing import Context
from htmy.utils import is_component_sequence

from .context import RendererContext

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable, Iterator

    from htmy.typing import Component, ComponentType, ContextProvider


class _Node:
    """A single node in the linked list the renderer constructs to resolve a component tree."""

    __slots__ = ("component", "next")

    def __init__(self, component: ComponentType, next: _Node | None = None) -> None:
        """
        Initialization.

        Arguments:
            component: The component in this node.
            next: The next component in the list, if there is one.
        """
        self.component = component
        self.next = next

    def iter_nodes(self, *, include_self: bool = True) -> Iterator[_Node]:
        """
        Iterates over all following nodes.

        Arguments:
            include_self: Whether the node on which this method is called should also
                be included in the iterator.
        """
        current = self if include_self else self.next
        while current is not None:
            yield current
            current = current.next


_NodeAndChildContext: TypeAlias = tuple[_Node, Context]


class _ComponentRenderer:
    """
    `ComponentType` renderer that converts a component tree into a linked list of resolved (`str`) nodes.
    """

    __slots__ = ("_async_todos", "_sync_todos", "_root", "_string_formatter")

    def __init__(
        self,
        component: ComponentType,
        context: Context,
        *,
        string_formatter: Callable[[str], str],
    ) -> None:
        """
        Initialization.

        Arguments:
            component: The component to render.
            context: The base context to use for rendering the component.
            string_formatter: The string formatter to use.
        """
        self._async_todos: deque[_NodeAndChildContext] = deque()
        """Async node - context tuples that need to be rendered."""
        self._sync_todos: deque[_NodeAndChildContext] = deque()
        """
        Sync node - context tuples that need to be rendered (`node.component` is an `HTMYComponentType`).
        """
        self._string_formatter = string_formatter
        """The string formatter to use."""

        if isinstance(component, str):
            root = _Node(string_formatter(component), None)
        else:
            root = _Node(component, None)
            self._schedule_node(root, context)
        self._root = root
        """The root node in the linked list the renderer constructs."""

    async def _extend_context(self, component: ContextProvider, context: Context) -> Context:
        """
        Returns a new context from the given component and context.

        Arguments:
            component: A `ContextProvider` component.
            context: The current rendering context.
        """
        extra_context: Context | Awaitable[Context] = component.htmy_context()
        if isawaitable(extra_context):
            extra_context = await extra_context

        return (
            # Context must not be mutated. We can ignore that ChainMap expects mutable mappings.
            ChainMap(extra_context, context)  # type: ignore[arg-type]
            if extra_context
            else context
        )

    def _process_node_result(self, parent_node: _Node, component: Component, context: Context) -> None:
        """
        Processes the result of a single node.

        Arguments:
            parent_node: The node that was resolved.
            component: The (awaited if async) result of `parent_node.component.htmy()`.
            context: The context that was used for rendering `parent_node.component`.
        """
        schedule_node = self._schedule_node
        string_formatter = self._string_formatter
        if hasattr(component, "htmy"):
            parent_node.component = component
            schedule_node(parent_node, context)
        elif isinstance(component, str):
            parent_node.component = string_formatter(component)
        elif is_component_sequence(component):
            if len(component) == 0:
                parent_node.component = ""
                return

            first_comp, *rest_comps = component
            if isinstance(first_comp, str):
                parent_node.component = string_formatter(first_comp)
            else:
                parent_node.component = first_comp
                schedule_node(parent_node, context)

            old_next = parent_node.next
            last: _Node = parent_node
            for c in rest_comps:
                if isinstance(c, str):
                    node = _Node(string_formatter(c), old_next)
                else:
                    node = _Node(c, old_next)
                    schedule_node(node, context)

                last.next = node
                last = node
        else:
            raise ValueError(f"Invalid component type: {type(component)}")

    async def _process_async_node(self, node: _Node, context: Context) -> None:
        """
        Processes the given node. `node.component` must be an async component.
        """
        result = await node.component.htmy(context)  # type: ignore[misc,union-attr]
        self._process_node_result(node, result, context)

    def _schedule_node(self, node: _Node, child_context: Context) -> None:
        """
        Schedules the given node for rendering with the given child context.

        `node.component` must be an `HTMYComponentType` (single component and not `str`).
        """
        component = node.component
        if component is None:
            pass  # Just skip the node
        elif iscoroutinefunction(component.htmy):  # type: ignore[union-attr]
            self._async_todos.append((node, child_context))
        else:
            self._sync_todos.append((node, child_context))

    async def run(self) -> str:
        """Runs the component renderer."""
        async_todos = self._async_todos
        sync_todos = self._sync_todos
        process_node_result = self._process_node_result
        process_async_node = self._process_async_node

        while sync_todos or async_todos:
            while sync_todos:
                node, child_context = sync_todos.pop()
                component = node.component
                if component is None:
                    continue

                if hasattr(component, "htmy_context"):  # isinstance() is too expensive.
                    child_context = await self._extend_context(component, child_context)  # type: ignore[arg-type]

                if iscoroutinefunction(component.htmy):  # type: ignore[union-attr]
                    async_todos.append((node, child_context))
                else:
                    result: Component = component.htmy(child_context)  # type: ignore[assignment,union-attr]
                    process_node_result(node, result, child_context)

            if async_todos:
                current_async_todos = async_todos
                self._async_todos = async_todos = deque()
                await asyncio_gather(*(process_async_node(n, ctx) for n, ctx in current_async_todos))

        return "".join(node.component for node in self._root.iter_nodes() if node.component is not None)  # type: ignore[misc]


async def _render_component(
    component: Component,
    *,
    context: Context,
    string_formatter: Callable[[str], str],
) -> str:
    """Renders the given component with the given settings."""
    if hasattr(component, "htmy"):
        return await _ComponentRenderer(component, context, string_formatter=string_formatter).run()
    elif isinstance(component, str):
        return string_formatter(component)
    elif is_component_sequence(component):
        if len(component) == 0:
            return ""

        renderers = (_ComponentRenderer(c, context, string_formatter=string_formatter) for c in component)
        return "".join(await asyncio_gather(*(r.run() for r in renderers)))
    elif component is None:
        return ""
    else:
        raise ValueError(f"Invalid component type: {type(component)}")


class Renderer:
    """
    The default renderer.

    It resolves component trees by converting them to a linked list of resolved component parts
    before combining them to the final string.
    """

    __slots__ = ("_default_context", "_string_formatter")

    def __init__(
        self,
        default_context: Context | None = None,
        *,
        string_formatter: Callable[[str], str] = xml_format_string,
    ) -> None:
        """
        Initialization.

        Arguments:
            default_context: The default context to use for rendering if `render()` doesn't
                receive a context.
            string_formatter: Callable that should be used to format plain strings. By default
                an XML-safe string formatter will be used.
        """
        self._default_context: Context = {} if default_context is None else default_context
        self._string_formatter = string_formatter

    async def render(self, component: Component, context: Context | None = None) -> str:
        """
        Renders the given component.

        Implements `htmy.typing.RendererType`.

        Arguments:
            component: The component to render.
            context: An optional rendering context.

        Returns:
            The rendered string.
        """
        # Create a new default context that also contains the renderer instance.
        # We must not put it in `self._default_context` because then the renderer
        # would keep a reference to itself.
        default_context = {**self._default_context, RendererContext: self}
        # Type ignore: ChainMap expects mutable mappings, but context mutation is not allowed so don't care.
        context = (
            default_context if context is None else ChainMap(context, default_context)  # type: ignore[arg-type]
        )
        return await _render_component(component, context=context, string_formatter=self._string_formatter)



from __future__ import annotations

from collections import ChainMap
from inspect import isawaitable
from typing import TYPE_CHECKING

from htmy.core import xml_format_string
from htmy.utils import is_component_sequence

from .context import RendererContext

if TYPE_CHECKING:
    from collections.abc import AsyncIterator, Awaitable, Callable

    from htmy.typing import Component, ComponentType, Context


class Renderer:
    """
    The baseline renderer that support both async streaming and rendering.

    Because of the simple, recursive implementation, this renderer is the easiest to reason about.
    Therefore it is useful for validating component correctness before bug reporting (if another
    renderer implementation fails), testing and debugging alternative implementations, and it can
    also serve as the baseline for benchmarking other renderer implementations.
    """

    __slots__ = ("_default_context", "_string_formatter")

    def __init__(
        self,
        default_context: Context | None = None,
        *,
        string_formatter: Callable[[str], str] = xml_format_string,
    ) -> None:
        """
        Initialization.

        Arguments:
            default_context: The default context to use for rendering if `render()` doesn't
                receive a context.
            string_formatter: Callable that should be used to format plain strings. By default
                an XML-safe string formatter will be used.
        """
        self._default_context: Context = {} if default_context is None else default_context
        self._string_formatter = string_formatter

    async def render(self, component: Component, context: Context | None = None) -> str:
        """
        Renders the given component.

        Implements `htmy.renderer.typing.RendererType`.

        Arguments:
            component: The component to render.
            context: An optional rendering context.

        Returns:
            The rendered string.
        """
        chunks = []
        async for chunk in self.stream(component, context):
            chunks.append(chunk)

        return "".join(chunks)

    async def stream(self, component: Component, context: Context | None = None) -> AsyncIterator[str]:
        """
        Async iterator that renders the given component.

        Implements `htmy.renderer.typing.StreamingRendererType`.

        Arguments:
            component: The component to render.
            context: An optional rendering context.

        Yields:
            The rendered strings.
        """
        # Create a new default context that also contains the renderer instance.
        # We must not put it in `self._default_context` because then the renderer
        # would keep a reference to itself.
        default_context = {**self._default_context, RendererContext: self}
        # Type ignore: ChainMap expects mutable mappings, but context mutation is not allowed so don't care.
        context = (
            default_context if context is None else ChainMap(context, default_context)  # type: ignore[arg-type]
        )

        async for chunk in self._stream(component, context):
            yield chunk

    async def _stream(self, component: Component, context: Context) -> AsyncIterator[str]:
        """
        Renders the given component with the given context.

        Arguments:
            component: The component to render.
            context: The current rendering context.

        Yields:
            String chunks of the rendered component.
        """
        if isinstance(component, str):
            yield self._string_formatter(component)
        elif component is None:
            return
        elif is_component_sequence(component):
            for comp in component:
                if comp is not None:
                    async for chunk in self._stream_one(comp, context):
                        yield chunk
        else:
            # Sync or async htmy component.
            async for chunk in self._stream_one(component, context):  # type: ignore[arg-type]
                yield chunk

    async def _stream_one(self, component: ComponentType, context: Context) -> AsyncIterator[str]:
        """
        Renders a single component.

        Arguments:
            component: The component to render.
            context: The current rendering context.

        Yields:
            The rendered strings.
        """
        if isinstance(component, str):
            yield self._string_formatter(component)
        elif component is None:
            return
        else:
            # Handle context providers
            child_context: Context = context
            if hasattr(component, "htmy_context"):  # isinstance() is too expensive.
                extra_context: Context | Awaitable[Context] = component.htmy_context()
                if isawaitable(extra_context):
                    extra_context = await extra_context

                if len(extra_context):
                    # Context must not be mutated, so we can ignore that ChainMap expects mutable mappings.
                    child_context = ChainMap(extra_context, context)  # type: ignore[arg-type]

            children = component.htmy(child_context)
            if isawaitable(children):
                children = await children

            async for chunk in self._stream(children, child_context):
                yield chunk



from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeAlias, TypedDict

from htmy.typing import Component

if TYPE_CHECKING:
    from typing_extensions import NotRequired
else:
    from typing import Optional as NotRequired

MarkdownMetadataDict: TypeAlias = dict[str, Any]


class ParsedMarkdown(TypedDict):
    """Type definition for parsed markdown data."""

    content: str
    metadata: NotRequired[MarkdownMetadataDict | None]


MarkdownParserFunction: TypeAlias = Callable[[str], ParsedMarkdown]
"""Callable that converts a markdown string into a `ParsedMarkdown` object."""

MarkdownRenderFunction: TypeAlias = Callable[[Component, MarkdownMetadataDict | None], Component]
"""Renderer function definition for markdown data."""



from __future__ import annotations

from collections.abc import AsyncIterator
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from typing import Any, TypeGuard

    from htmy.typing import Component, Context


class RendererType(Protocol):
    """Protocol definition for renderers."""

    async def render(self, component: Component, context: Context | None = None) -> str:
        """
        Renders the given component.

        Arguments:
            component: The component to render.
            context: An optional rendering context.

        Returns:
            The rendered string.
        """
        ...


def is_renderer(obj: Any | None) -> TypeGuard[RendererType]:
    """Type guard that checks if the given object is a renderer."""
    # Just a basic check, don't waste time here.
    render: Any = getattr(obj, "render", None)
    return render is not None


class StreamingRendererType(RendererType, Protocol):
    """Protocol definition for streaming renderers."""

    def stream(self, component: Component, context: Context | None = None) -> AsyncIterator[str]:
        """
        Async iterator that renders the given component.

        Arguments:
            component: The component to render.
            context: An optional rendering context.

        Yields:
            The rendered strings.
        """
        ...


def is_streaming_renderer(obj: Any | None) -> TypeGuard[StreamingRendererType]:
    """Type guard that checks if the given object is a streaming renderer."""
    if not is_renderer(obj):
        return False

    # Just a basic check, don't waste time here.
    stream: Any = getattr(obj, "stream", None)
    return stream is not None



from __future__ import annotations

from typing import TYPE_CHECKING

from .core import SafeStr
from .renderer.context import RendererContext
from .utils import as_component_type

if TYPE_CHECKING:
    from collections.abc import Container

    from .typing import Component, ComponentType, Context


class ErrorBoundary:
    """
    Error boundary component for graceful error handling.

    If an error occurs during the rendering of the error boundary's subtree,
    the fallback component will be rendered instead.
    """

    __slots__ = ("_children", "_errors", "_fallback")

    def __init__(
        self,
        *children: ComponentType,
        fallback: Component = None,
        errors: Container[type[Exception]] | None = None,
    ) -> None:
        """
        Initialization.

        Arguments:
            *children: The wrapped children components.
            fallback: The fallback component to render in case an error occurs during children rendering.
            errors: An optional set of accepted error types. Only accepted errors are swallowed and rendered
                with the fallback. If an error is not in this set but one of its base classes is, then the
                error will still be accepted and the fallback rendered. By default all errors are accepted.
        """
        self._children = children
        self._errors = errors
        self._fallback: Component = fallback

    async def htmy(self, context: Context) -> Component:
        renderer = RendererContext.from_context(context)

        try:
            result = await renderer.render(self._children, context)
        except Exception as e:
            result = await renderer.render(self._fallback_component(e), context)

        # We must convert the already rendered string to SafeStr to avoid additional XML escaping.
        return SafeStr(result)

    def _fallback_component(self, error: Exception) -> ComponentType:
        """
        Returns the fallback component for the given error.

        Arguments:
            error: The error that occurred during the rendering of the error boundary's subtree.

        Raises:
            Exception: The received error if it's not accepted.
        """
        if not (self._errors is None or any(e in self._errors for e in type(error).mro())):
            raise error

        return as_component_type(self._fallback)



from htmy.typing import Context

from .typing import RendererType, is_renderer


class RendererContext:
    """
    Context key for storing, and utility for retrieving the renderer instance from the rendering context.
    """

    @classmethod
    def from_context(cls, context: Context) -> RendererType:
        """
        Returns the renderer instance from the given context.

        Arguments:
            context: The context the renderer instance should be loaded from.

        Raises:
            KeyError: If no renderer instance was found in the context.
            TypeError: If the value corresponding to `RendererContext` in the context is not a renderer.
        """
        renderer = context[cls]

        if not is_renderer(renderer):
            raise TypeError("The value corresponding to RendererContext in the context must be a renderer.")

        return renderer



from __future__ import annotations

import re
from collections.abc import Iterator, Mapping
from inspect import isawaitable
from typing import TYPE_CHECKING

from async_lru import alru_cache

from .core import SafeStr, Text
from .io import load_text_file
from .typing import (
    Component,
    ComponentType,
    Context,
    TextProcessor,
    TextResolver,
)
from .utils import as_component_sequence, as_component_type, is_component_sequence

if TYPE_CHECKING:
    from pathlib import Path

# -- Components and utilities


class Slots:
    """
    Utility that resolves slots in a string input to components.

    More technically, it splits a string into slot and non-slot parts, replaces the
    slot parts with the corresponding components (which may be component sequences)
    from the given slot mapping, and returns the resulting component sequence.

    The default slot placeholder is a standard XML/HTML comment of the following form:
    `<!-- slot[slot-key] -->`. Any number of whitespaces (including 0) are allowed in
    the placeholder, but the slot key must not contain any whitespaces. For details, see
    `Slots.slot_re`.

    Besides the pre-defined regular expressions in `Slots.slot_re`, any other regular
    expression can be used to identify slots as long as it meets the requirements described
    in `Slots.slots_re`.

    Implements: `htmy.typing.TextResolver`
    """

    __slots__ = ("_not_found", "_slot_mapping", "_slot_re")

    class slot_re:
        """
        Slot regular expressions.

        Requirements:

        - The regular expression must have exactly one capturing group that captures the slot key.
        """

        square_bracket = re.compile(r"<!-- *slot *\[ *([^[ ]+) *\] *-->")
        """
        Slot regular expression that matches slots defined as follows: `<!-- slot[slot-key] -->`.

        The slot key must not contain any whitespaces and there must not be any additional text
        in the XML/HTML comment. Any number of whitespaces (including 0) are allowed around the
        parts of the slot placeholder.
        """
        parentheses = re.compile(r"<!-- *slot *\( *([^( ]+) *\) *-->")
        """
        Slot regular expression that matches slots defined as follows: `<!-- slot(slot-key) -->`.

        The slot key must not contain any whitespaces and there must not be any additional text
        in the XML/HTML comment. Any number of whitespaces (including 0) are allowed around the
        parts of the slot placeholder.
        """

        # There are no defaults for angle bracket and curly braces, because
        # they may conflict with HTML and format strings.

        default = square_bracket
        """
        The default slot regular expression. Same as `Slots.slot_re.square_bracket`.
        """

    def __init__(
        self,
        slot_mapping: Mapping[str, Component],
        *,
        slot_re: re.Pattern[str] = slot_re.default,
        not_found: Component | None = None,
    ) -> None:
        """
        Initialization.

        Slot regular expressions are used to find slot keys in strings, which are then replaced
        with the corresponding component from the slot mapping. `slot_re` must have exactly one
        capturing group that captures the slot key. `Slots.slot_re` contains some predefined slot
        regular expressions, but any other regular expression can be used as long as it matches
        the capturing group requirement above.

        Arguments:
            slot_mapping: Slot mapping the maps slot keys to the corresponding component.
            slot_re: The slot regular expression that is used to find slot keys in strings.
            not_found: The component that is used to replace slot keys that are not found in
                `slot_mapping`. If `None` and the slot key is not found in `slot_mapping`,
                then a `KeyError` will be raised by `resolve()`.
        """
        self._slot_mapping = slot_mapping
        self._slot_re = slot_re
        self._not_found = not_found

    def resolve_text(self, text: str) -> Component:
        """
        Resolves the given string into components using the instance's slot regular expression
        and slot mapping.

        Arguments:
            text: The text to resolve.

        Returns:
           The component sequence the text resolves to.

        Raises:
            KeyError: If a slot key is not found in the slot mapping and `not_found` is `None`.
        """
        return tuple(self._resolve_text(text))

    def _resolve_text(self, text: str) -> Iterator[ComponentType]:
        """
        Generator that yields the slot and non-slot parts of the given string in order.

        Arguments:
            text: The text to resolve.

        Yields:
            The slot and non-slot parts of the given string.

        Raises:
            KeyError: If a slot key is not found in the slot mapping and `not_found` is `None`.
        """
        is_slot = False
        # The implementation requires that the slot regular expression has exactly one capturing group.
        for part in self._slot_re.split(text):
            if is_slot:
                resolved = self._slot_mapping.get(part, self._not_found)
                if resolved is None:
                    raise KeyError(f"Component not found for slot: {part}")

                if is_component_sequence(resolved):
                    yield from resolved
                else:
                    # mypy complains that resolved may be a sequence, but that's not the case.
                    yield resolved  # type: ignore[misc]
            else:
                yield part

            is_slot = not is_slot


class Snippet:
    """
    Component that renders text, which may be asynchronously loaded from a file.

    **Warning:** The component treats its input as trusted. If any part of the input comes from
    untrusted sources, ensure it is safely escaped (using for example `htmy.xml_format_string`)!
    Passing untrusted input to this component leads to XSS vulnerabilities.

    The entire snippet processing pipeline consists of the following steps:

    1. The text content is loaded from a file or passed directly as a `Text` instance.
    2. The text content is processed by a `TextProcessor` if provided.
    3. The processed text is converted into a component (may be component sequence)
       by a `TextResolver`, for example `Slots`.
    4. Every `str` children (produced by the steps above) is converted into a `SafeStr` for
       rendering.

    The pipeline above is a bit abstract, so here are some usage notes:

    - The text content of a snippet can be a Python format string template, in which case the
      `TextProcessor` can be a simple method that calls `str.format()` with the correct arguments.
    - Alternatively, a text processor can also be used to get only a substring -- commonly referred
      to as fragment in frameworks like Jinja -- of the original text.
    - The text processor is applied before the text resolver, which makes it possible to insert
      placeholders into the text (for example slots, like in this case:
      `..."{toolbar}...".format(toolbar="<!-- slot[toolbar] -->")`) that are then replaced with any
      `htmy.Component` by the `TextResolver` (for example `Slots`).
    - `TextResolver` can return plain `str` values, it is not necessary for it to convert strings
      to `SafeStr` to prevent unwanted escaping.

    Example:

    ```python
    from datetime import date
    from htmy import Snippet, Slots

    def text_processor(text: str, context: Context) -> str:
       return text.format(today=date.today())

    snippet = Snippet(
        "my-page.html",
        text_processor=text_processor,
        text_resolver=Slots(
            {
                "date-picker": MyDatePicker(class_="text-primary"),
                "Toolbar": MyPageToolbar(active_page="home"),
                ...
            }
        ),
    )
    ```

    In the above example, if `my-page.html` contains a `{today}` placeholder, it will be replaced
    with the current date. If it contains a `<!-- slot[toolbar] -->}` slot, then the `MyPageToolbar`
    `htmy` component instance will be rendered in its place, and the `<!-- slot[date-picker] -->` slot
    will be replaced with the `MyDatePicker` component instance.
    """

    __slots__ = ("_path_or_text", "_text_processor", "_text_resolver")

    def __init__(
        self,
        path_or_text: Text | str | Path,
        text_resolver: TextResolver | None = None,
        *,
        text_processor: TextProcessor | None = None,
    ) -> None:
        """
        Initialization.

        Arguments:
            path_or_text: The path from where the content should be loaded or a `Text`
                instance if this value should be rendered directly.
            text_resolver: An optional `TextResolver` (e.g. `Slots`) that converts the processed
                text into a component. If not provided, the text will be rendered as a `SafeStr`.
            text_processor: An optional `TextProcessor` that can be used to process the text
                content before rendering. It can be used for example for token replacement or
                string formatting.
        """
        self._path_or_text = path_or_text
        self._text_processor = text_processor
        self._text_resolver = text_resolver

    async def htmy(self, context: Context) -> Component:
        """Renders the component."""
        text = await self._get_text_content()
        if self._text_processor is not None:
            processed = self._text_processor(text, context)
            text = (await processed) if isawaitable(processed) else processed

        if self._text_resolver is None:
            return self._render_text(text, context)

        comps = as_component_sequence(self._text_resolver.resolve_text(text))
        return tuple(
            as_component_type(self._render_text(c, context)) if isinstance(c, str) else c for c in comps
        )

    async def _get_text_content(self) -> str:
        """Returns the plain text content that should be rendered."""
        path_or_text = self._path_or_text

        if isinstance(path_or_text, Text):
            return path_or_text
        else:
            return await Snippet._load_text_file(path_or_text)

    def _render_text(self, text: str, context: Context) -> Component:
        """
        Render function that takes the text that must be rendered and the current rendering context,
        and returns the corresponding component.
        """
        return SafeStr(text)

    @staticmethod
    @alru_cache()
    async def _load_text_file(path: str | Path) -> str:
        """Async text loader with an LRU cache."""
        return await load_text_file(path)



from __future__ import annotations

import enum
import json
from typing import TYPE_CHECKING, Any, ClassVar, cast
from xml.sax.saxutils import escape as xml_escape
from xml.sax.saxutils import quoteattr as xml_quoteattr

if TYPE_CHECKING:
    from collections.abc import Callable

    from typing_extensions import Never, Self

    from .typing import Component, ComponentType, Context, ContextKey, ContextValue, T

# -- Utility components


class Fragment:
    """Fragment utility component that simply wraps some children components."""

    __slots__ = ("_children",)

    def __init__(self, *children: ComponentType) -> None:
        """
        Initialization.

        Arguments:
            *children: The wrapped children.
        """
        self._children = children

    def htmy(self, context: Context) -> Component:
        """Renders the component."""
        return self._children


class WithContext(Fragment):
    """
    A simple, static context provider component.
    """

    __slots__ = ("_context",)

    def __init__(self, *children: ComponentType, context: Context) -> None:
        """
        Initialization.

        Arguments:
            *children: The children components to wrap in the given context.
            context: The context to make available to children components.
        """
        super().__init__(*children)
        self._context = context

    def htmy_context(self) -> Context:
        """Returns the context for child rendering."""
        return self._context


# -- Context utilities


class ContextAware:
    """
    Base class with utilities for safe context use.

    Features:

    - Register subclass instance in a context.
    - Load subclass instance from context.
    - Wrap components within a subclass instance context.

    Subclass instance registration:

    Direct subclasses are considered the "base context type". Subclass instances are
    registered in contexts under their own type and also under their "base context type".

    Example:

    ```python
    class ContextDataDefinition(ContextAware):
        # This is the "base context type", instances of this class and its subclasses
        # will always be registered under this type.
        ...

    class ContextDataImplementation(ContextDataDefinition):
        # Instances of this class will be registered under `ContextDataDefinition` (the
        # "base context type") and also under this type.
        ...

    class SpecializedContextDataImplementation(ContextDataImplementation):
        # Instances of this class will be registered under `ContextDataDefinition` (the
        # "base context type") and also under this type, but they will not be registered
        # under `ContextDataImplementation`, since that's not the base context type.
        ...
    ```
    """

    __slots__ = ()

    _base_context_type: ClassVar[type[ContextAware] | None] = None

    def __init_subclass__(cls) -> None:
        if cls.mro()[1] == ContextAware:
            cls._base_context_type = cls

    def in_context(self, *children: ComponentType) -> WithContext:
        """
        Creates a context provider component that renders the given children using this
        instance in its context.
        """
        return WithContext(*children, context=self.to_context())

    def to_context(self) -> Context:
        """
        Creates a context with this instance in it.

        See the context registration rules in the class documentation for more information.
        """
        result: dict[ContextKey, ContextValue] = {type(self): self}
        if self._base_context_type is not None:
            result[self._base_context_type] = self

        return result

    @classmethod
    def from_context(cls, context: Context, default: Self | None = None) -> Self:
        """
        Looks up an instance of this class from the given contexts.

        Arguments:
            context: The context the instance should be loaded from.
            default: The default to use if no instance was found in the context.
        """
        result = context[cls] if default is None else context.get(cls, default)
        if isinstance(result, cls):
            return result

        raise TypeError(f"Invalid context data type for {cls.__name__}.")


# -- Formatting


class SkipProperty(Exception):
    """Exception raised by property formatters if the property should be skipped."""

    ...

    @classmethod
    def format_property(cls, _: Any) -> Never:
        """Property formatter that raises a `SkipProperty` error regardless of the received value."""
        raise cls("skip-property")


class Text(str):
    """Marker class for differentiating text content from other strings."""

    ...


class SafeStr(Text):
    """
    String subclass whose instances shouldn't get escaped during rendering.

    Note: any operation on `SafeStr` instances will result in plain `str` instances which
    will be rendered normally. Make sure the `str` to `SafeStr` conversion (`SafeStr(my_string)`)
    takes when there's no string operation afterwards.
    """

    ...


class XBool(enum.Enum):
    """
    Utility for the valid formatting of boolean XML (and HTML) attributes.

    See this article for more information:
    https://developer.mozilla.org/en-US/docs/Web/HTML/Attributes#boolean_attributes
    """

    true = True
    false = False

    def format(self) -> str:
        """
        Raises `SkipProperty` for `XBool.false`, returns empty string for `XBool.true`.
        """
        if self is XBool.true:
            return ""

        raise SkipProperty()


def xml_format_string(value: str) -> str:
    """Escapes `<`, `>`, and `&` characters in the given string, unless it's a `SafeStr`."""
    return value if isinstance(value, SafeStr) else xml_escape(value)


class Formatter(ContextAware):
    """
    The default, context-aware property name and value formatter.

    The formatter supports both primitive and (many) complex values, such as lists,
    dictionaries, tuples, and sets. Complex values are JSON-serialized by default.

    Important: the default implementation looks up the formatter for a given value by checking
    its type, but it doesn't do this check with the base classes of the encountered type. For
    example the formatter will know how to format `datetime` object, but it won't know how to
    format a `MyCustomDatetime(datetime)` instance.

    One reason for this is efficiency: always checking the base classes of every single value is a
    lot of unnecessary calculation. The other reason is customizability: this way you could use
    subclassing for formatter selection, e.g. with `LocaleDatetime(datetime)`-like classes.

    Property name and value formatters may raise a `SkipProperty` error if a property should be skipped.
    """

    __slots__ = ("_default_formatter", "_name_formatter", "_value_formatters")

    def __init__(
        self,
        *,
        default_formatter: Callable[[Any], str] = str,
        name_formatter: Callable[[str], str] | None = None,
    ) -> None:
        """
        Initialization.

        Arguments:
            default_formatter: The default property value formatter to use if no formatter could
                be found for a given value.
            name_formatter: Optional property name formatter (for replacing the default name formatter).
        """
        super().__init__()
        self._default_formatter = default_formatter
        self._name_formatter = self._format_name if name_formatter is None else name_formatter
        self._value_formatters: dict[type, Callable[[Any], str]] = self._base_formatters()

    def add(self, key: type[T], formatter: Callable[[T], str]) -> Self:
        """Registers the given value formatter under the given key."""
        self._value_formatters[key] = formatter
        return self

    def format(self, name: str, value: Any) -> str:
        """
        Formats the given name-value pair.

        Returns an empty string if the property name or value should be skipped.

        See `SkipProperty` for more information.
        """
        try:
            return f"{self.format_name(name)}={xml_quoteattr(self.format_value(value))}"
        except SkipProperty:
            return ""

    def format_name(self, name: str) -> str:
        """
        Formats the given name.

        Raises:
            SkipProperty: If the property should be skipped.
        """
        return self._name_formatter(name)

    def format_value(self, value: Any) -> str:
        """
        Formats the given value.

        Arguments:
            value: The property value to format.

        Raises:
            SkipProperty: If the property should be skipped.
        """
        fmt = self._value_formatters.get(type(value), self._default_formatter)
        return fmt(value)

    def _format_name(self, name: str, /) -> str:
        """The default property name formatter."""
        no_replacement = "_" in {name[0], name[-1]}
        return name.strip("_") if no_replacement else name.replace("_", "-")

    def _base_formatters(self) -> dict[type, Callable[[Any], str]]:
        """Factory that creates the default value formatter mapping."""
        from datetime import date, datetime

        return {
            bool: lambda v: "true" if v else "false",
            date: lambda d: cast(date, d).isoformat(),
            datetime: lambda d: cast(datetime, d).isoformat(),
            dict: lambda v: json.dumps(v),
            list: lambda v: json.dumps(v),
            tuple: lambda v: json.dumps(v),
            set: lambda v: json.dumps(tuple(v)),
            XBool: lambda v: cast(XBool, v).format(),
            type(None): SkipProperty.format_property,
        }



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from markdown import Markdown

from htmy.core import ContextAware, SafeStr, Text
from htmy.snippet import Snippet
from htmy.typing import TextProcessor, TextResolver

if TYPE_CHECKING:
    from collections.abc import Callable
    from pathlib import Path

    from htmy.typing import Component, Context

    from .typing import MarkdownParserFunction, MarkdownRenderFunction, ParsedMarkdown


class MarkdownParser(ContextAware):
    """
    Context-aware markdown parser.

    By default, this class uses the `markdown` library with a sensible set of
    [extensions](https://python-markdown.github.io/extensions/) including code highlighting.
    """

    __slots__ = ("_md",)

    _default: ClassVar[MarkdownParser | None] = None
    """The default instance or `None` if one hasn't been created already."""

    @classmethod
    def default(cls) -> MarkdownParser:
        """
        Returns the default instance.
        """
        if cls._default is None:
            cls._default = MarkdownParser()

        return cls._default

    def __init__(self, md: MarkdownParserFunction | None = None) -> None:
        """
        Initialization.

        Arguments:
            md: The parser function to use.
        """
        super().__init__()
        self._md = md

    def parse(self, text: str) -> ParsedMarkdown:
        """
        Returns the markdown data by parsing the given text.
        """
        md = self._md
        if md is None:
            md = self._default_md()
            self._md = md

        return md(text)

    def _default_md(self) -> MarkdownParserFunction:
        """
        Function that creates the default markdown parser.

        Returns:
            The default parser function.
        """
        md = Markdown(extensions=("extra", "meta", "codehilite"))

        def parse(text: str) -> ParsedMarkdown:
            md.reset()
            parsed = md.convert(text)
            return {"content": parsed, "metadata": getattr(md, "Meta", None)}

        return parse


class MD(Snippet):
    """
    Component for reading, customizing, and rendering markdown documents.

    It supports all the processing utilities of `Snippet`, including `text_resolver` and
    `text_processor` for formatting, token replacement, and slot conversion to components.

    One note regarding slot conversion (`text_resolver`): it is executed before markdown parsing,
    and all string segments of the resulting component sequence are parsed individually by the
    markdown parser. As a consequence, you should only use slots in places where the preceding
    and following texts individually result in valid markdown.

    **Warning:** The component treats its input as trusted. If any part of the input comes from
    untrusted sources, ensure it is safely escaped (using for example `htmy.xml_format_string`)!
    Passing untrusted input to this component leads to XSS vulnerabilities.
    """

    __slots__ = (
        "_converter",
        "_renderer",
    )

    def __init__(
        self,
        path_or_text: Text | str | Path,
        text_resolver: TextResolver | None = None,
        *,
        converter: Callable[[str], Component] | None = None,
        renderer: MarkdownRenderFunction | None = None,
        text_processor: TextProcessor | None = None,
    ) -> None:
        """
        Initialization.

        Arguments:
            path_or_text: The path where the markdown file is located or a markdown `Text`.
            text_resolver: An optional `TextResolver` (e.g. `Slots`) that converts the processed
                text into a component.
            converter: Function that converts an HTML string (the parsed and processed markdown text)
                into a component.
            renderer: Function that gets the parsed and converted content and the metadata (if it exists)
                and turns them into a component.
            text_processor: An optional text processors that can be used to process the text
                content before rendering. It can be used for example for token replacement or
                string formatting.
        """
        super().__init__(path_or_text, text_resolver, text_processor=text_processor)
        self._converter: Callable[[str], Component] = SafeStr if converter is None else converter
        self._renderer = renderer

    def _render_text(self, text: str, context: Context) -> Component:
        md = MarkdownParser.from_context(context, MarkdownParser.default()).parse(text)
        result = self._converter(md["content"])
        return result if self._renderer is None else self._renderer(result, md.get("metadata", None))



from collections.abc import Callable, Coroutine, Mapping, MutableMapping
from typing import Any, Protocol, TypeAlias, TypeVar

T = TypeVar("T")
U = TypeVar("U")

# -- Properties

PropertyValue: TypeAlias = Any | None
"""Component/XML tag property value."""

Properties: TypeAlias = Mapping[str, PropertyValue]
"""Component/XML tag property mapping."""

# -- Context

ContextKey: TypeAlias = Any
"""Context key."""

ContextValue: TypeAlias = Any
"""Context value."""

Context: TypeAlias = Mapping[ContextKey, ContextValue]
"""Context mapping."""

MutableContext: TypeAlias = MutableMapping[ContextKey, ContextValue]
"""
Mutable context mapping.

It can be helpful when the created context should be marked as mutable for static type analysis
(usually the created context is a plain `dict`).
"""

# -- Components


class SyncComponent(Protocol):
    """Protocol definition for sync `htmy` components."""

    def htmy(self, context: Context, /) -> "Component":
        """Renders the component."""
        ...


class AsyncComponent(Protocol):
    """Protocol definition for async `htmy` components."""

    async def htmy(self, context: Context, /) -> "Component":
        """Renders the component."""
        ...


HTMYComponentType: TypeAlias = SyncComponent | AsyncComponent
"""Sync or async `htmy` component type."""

StrictComponentType: TypeAlias = HTMYComponentType | str
"""Type definition for a single component that's not `None`."""

ComponentType: TypeAlias = StrictComponentType | None
"""Type definition for a single component."""

# Omit strings from this type to simplify checks.
ComponentSequence: TypeAlias = list[ComponentType] | tuple[ComponentType, ...]
"""Component sequence type."""

Component: TypeAlias = ComponentType | ComponentSequence
"""Component type: a single component or a sequence of components."""


# -- Context providers


class SyncContextProvider(Protocol):
    """Protocol definition for sync context providers."""

    def htmy_context(self) -> Context:
        """Returns a context for child rendering."""
        ...


class AsyncContextProvider(Protocol):
    """Protocol definition for async context providers."""

    async def htmy_context(self) -> Context:
        """Returns a context for child rendering."""
        ...


ContextProvider: TypeAlias = SyncContextProvider | AsyncContextProvider
"""
Sync or async context provider type.

Components can implement this protocol to add extra data to the rendering context
of their entire component subtree (including themselves).
"""

# -- Text processors

TextProcessor: TypeAlias = Callable[[str, Context], str | Coroutine[Any, Any, str]]
"""Callable type that expects a string and a context, and returns a processed string."""


class TextResolver(Protocol):
    """
    Protocol definition for resolvers that convert a string to a component.
    """

    def resolve_text(self, text: str) -> Component:
        """
        Returns the resolved component for the given text.

        Arguments:
            text: The text to resolve.

        Raises:
            KeyError: If the text cannot be resolved to a component.
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .typing import ComponentType


from .core import SafeStr
from .tag import Tag, TagWithProps


class DOCTYPE:
    """Document type declaration."""

    html = SafeStr("<!DOCTYPE html>")
    """HTML document type."""


html = Tag("html")
"""
`<html>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/html.
"""

head = Tag("head")
"""
`<head>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/head.
"""

body = Tag("body")
"""
`<body>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/body.
"""

base = TagWithProps("base")
"""
`<base>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/base.
"""

title = Tag("title", child_separator=None)  # TODO: allow only a single str child.
"""
`<title>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/title.
"""

link = TagWithProps("link")
"""
`<link>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/link.
"""


class Link:
    """`link` tag factories."""

    @staticmethod
    def css(href: str) -> ComponentType:
        return link(rel="stylesheet", type="text/css", href=href)


meta = TagWithProps("meta")
"""
`<meta>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meta.
"""


class Meta:
    """`meta` tag factories."""

    @staticmethod
    def author(content: str) -> ComponentType:
        return meta(name="author", content=content)

    @staticmethod
    def charset(charset: str = "utf-8") -> ComponentType:
        return meta(charset=charset)

    @staticmethod
    def description(content: str) -> ComponentType:
        return meta(name="description", content=content)

    @staticmethod
    def keywords(content: str) -> ComponentType:
        return meta(name="keywords", content=content)

    @staticmethod
    def viewport(content: str = "width=device-width, initial-scale=1.0") -> ComponentType:
        return meta(name="viewport", content=content)


script = Tag("script")  # TODO: allow only a single str child.
"""
`<script>` element.

The script tag should have only one child, the script content. If it's not
empty, it should be a `SafeStr` for HTML and plain `str` for XHTML.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/script.
"""

style = Tag("style")  # TODO: only allow a single str child, automatically convert plain str to SafeStr.
"""
`<style>` element.

The style tag should have only one child, the style content. If it's not
empty, it should be a `SafeStr` to avoid XML escaping.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/style.
"""

dialog = Tag("dialog")
"""
`<dialog>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/dialog.
"""

address = Tag("address")
"""
`<address>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/address.
"""

article = Tag("article")
"""
`<article>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/article.
"""

aside = Tag("aside")
"""
`<aside>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/aside.
"""

blockquote = Tag("blockquote")
"""
`<blockquote>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/blockquote.
"""

div = Tag("div")
"""
`<div>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/div.
"""

embed = TagWithProps("embed")
"""
`<embed>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/embed.
"""

figure = Tag("figure")
"""
`<figure>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/figure.
"""

figcaption = Tag("figcaption")
"""
`<figcaption>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/figcaption.
"""

footer = Tag("footer")
"""
`<footer>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/footer.
"""

header = Tag("header")
"""
`<header>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/header.
"""

hgroup = Tag("hgroup")
"""
`<hgroup>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/hgroup.
"""

hr = TagWithProps("hr")
"""
`<hr>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/hr.
"""

iframe = Tag("iframe")
"""
`<iframe>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/iframe.
"""

main = Tag("main")
"""
`<main>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/main.
"""

details = Tag("details")
"""
`<details>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/details.
"""

summary = Tag("summary", child_separator=None)
"""
`<summary>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/summary.
"""

nav = Tag("nav")
"""
`<nav>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/nav.
"""

menu = Tag("menu")
"""
`<menu>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/menu.
"""

noscript = Tag("noscript")
"""
`<noscript>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/noscript.
"""

pre = Tag("pre", child_separator=None)
"""
`<pre>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/pre.
"""

section = Tag("section")
"""
`<section>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/section.
"""

template = Tag("template")
"""
`<template>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/template.
"""

form = Tag("form")
"""
`<form>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/form.
"""

search = Tag("search")
"""
`<search>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/search.
"""

button = Tag("button")
"""
`<button>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/button.
"""

option = Tag("option")
"""
`<option>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/option.
"""

optgroup = Tag("optgroup")
"""
`<optgroup>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/optgroup.
"""

datalist = Tag("datalist")
"""
`<datalist>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/datalist.
"""

fieldset = Tag("fieldset")
"""
`<fieldset>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/fieldset.
"""

input_ = TagWithProps("input")
"""
`<input>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/input.
"""

label = Tag("label")
"""
`<label>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/label.
"""

legend = Tag("legend")
"""
`<legend>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/legend.
"""

meter = Tag("meter")
"""
`<meter>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/meter.
"""

object = Tag("object")
"""
`<object>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/object.
"""

output = Tag("output")
"""
`<output>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/output.
"""

progress = Tag("progress")
"""
`<progress>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/progress.
"""

select = Tag("select")
"""
`<select>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/select.
"""

textarea = Tag("textarea")
"""
`<textarea>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/textarea.
"""

a = Tag("a", child_separator=None)
"""
`<a>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/a.
"""

abbr = Tag("abbr", child_separator=None)
"""
`<abbr>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/abbr.
"""

b = Tag("b", child_separator=None)
"""
`<b>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/b.
"""

bdi = Tag("bdi", child_separator=None)
"""
`<bdi>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/bdi.
"""

bdo = Tag("bdo", child_separator=None)
"""
`<bdo>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/bdo.
"""

br = TagWithProps("br")
"""
`<br>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/br.
"""

cite = Tag("cite", child_separator=None)
"""
`<cite>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/cite.
"""

code = Tag("code", child_separator=None)
"""
`<code>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/code.
"""

data = Tag("data", child_separator=None)
"""
`<data>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/data.
"""

del_ = Tag("del", child_separator=None)
"""
`<del>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/del.
"""

dfn = Tag("dfn", child_separator=None)
"""
`<dfn>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/dfn.
"""

em = Tag("em", child_separator=None)
"""
`<em>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/em.
"""

i = Tag("i", child_separator=None)
"""
`<i>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/i.
"""

kbd = Tag("kbd", child_separator=None)
"""
`<kbd>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/kbd.
"""

picture = Tag("picture")
"""
`<picture>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/picture.
"""

img = TagWithProps("img")
"""
`<img>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/img.
"""

source = TagWithProps("source")
"""
`<source>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/source.
"""

ins = Tag("ins", child_separator=None)
"""
`<ins>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/ins.
"""

mark = Tag("mark", child_separator=None)
"""
`<mark>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/mark.
"""

q = Tag("q", child_separator=None)
"""
`<q>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/q.
"""

s = Tag("s", child_separator=None)
"""
`<s>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/s.
"""

samp = Tag("samp", child_separator=None)
"""
`<samp>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/samp.
"""

small = Tag("small", child_separator=None)
"""
`<small>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/small.
"""

span = Tag("span", child_separator=None)
"""
`<span>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/span.
"""

strong = Tag("strong", child_separator=None)
"""
`<strong>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/strong.
"""

sub = Tag("sub", child_separator=None)
"""
`<sub>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/sub.
"""

sup = Tag("sup", child_separator=None)
"""
`<sup>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/sup.
"""

svg = Tag("svg", child_separator=None)
"""
`<svg>` element.

See https://developer.mozilla.org/en-US/docs/Web/SVG/Element/svg.
"""

u = Tag("u", child_separator=None)
"""
`<u>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/u.
"""

var = Tag("var", child_separator=None)
"""
`<var>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/var.
"""

wbr = TagWithProps("wbr")
"""
`<wbr>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/wbr.
"""

li = Tag("li", child_separator=None)
"""
`<li>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/li.
"""

ol = Tag("ol")
"""
`<ol>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/ol.
"""

ul = Tag("ul")
"""
`<ul>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/ul.
"""

dl = Tag("dl")
"""
`<dl>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/dl.
"""

dt = Tag("dt", child_separator=None)
"""
`<dt>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/dt.
"""

dd = Tag("dd")
"""
`<dd>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/dd.
"""

caption = Tag("caption")
"""
`<caption>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/caption.
"""

table = Tag("table")
"""
`<table>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/table.
"""

thead = Tag("thead")
"""
`<thead>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/thead.
"""

tbody = Tag("tbody")
"""
`<tbody>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/tbody.
"""

tfoot = Tag("tfoot")
"""
`<tfoot>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/tfoot.
"""

tr = Tag("tr")
"""
`<tr>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/tr.
"""

th = Tag("th", child_separator=None)
"""
`<th>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/th.
"""

td = Tag("td", child_separator=None)
"""
`<td>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/td.
"""

colgroup = Tag("colgroup")
"""
`<colgroup>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/colgroup.
"""

col = TagWithProps("col")
"""
`<col>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/col.
"""

h1 = Tag("h1", child_separator=None)
"""
`<h1>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h1.
"""

h2 = Tag("h2", child_separator=None)
"""
`<h2>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h2.
"""

h3 = Tag("h3", child_separator=None)
"""
`<h3>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h3.
"""

h4 = Tag("h4", child_separator=None)
"""
`<h4>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h4.
"""

h5 = Tag("h5", child_separator=None)
"""
`<h5>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h5.
"""

h6 = Tag("h6", child_separator=None)
"""
`<h6>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/h6.
"""

p = Tag("p", child_separator=None)
"""
`<p>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/p.
"""

time = Tag("time", child_separator=None)
"""
`<time>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/time.
"""

audio = Tag("audio")
"""
`<audio>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/audio.
"""

video = Tag("video")
"""
`<video>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/video.
"""

track = TagWithProps("track")
"""
`<track>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/track.
"""

canvas = Tag("canvas")
"""
`<canvas>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/canvas.
"""

area = TagWithProps("area")
"""
`<area>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/area.
"""

map = Tag("map")
"""
`<map>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/map.
"""

slot = Tag("slot")
"""
`<slot>` element.

See https://developer.mozilla.org/en-US/docs/Web/HTML/Element/slot.
"""


class entity:
    amp = SafeStr("&amp;")
    apos = SafeStr("&apos;")
    cent = SafeStr("&cent;")
    copy = SafeStr("&copy;")
    euro = SafeStr("&euro;")
    gt = SafeStr("&gt;")
    lt = SafeStr("&lt;")
    mdash = SafeStr("&mdash;")
    nbsp = SafeStr("&nbsp;")
    ndash = SafeStr("&ndash;")
    pound = SafeStr("&pound;")
    quot = SafeStr("&quot;")
    reg = SafeStr("&reg;")
    times = SafeStr("&times;")
    yen = SafeStr("&yen;")
